#![no_std]
#![no_main]
#![feature(type_alias_impl_trait, async_fn_in_trait)]

use core::cell::Cell;
use cortex_m::prelude::{
    _embedded_hal_blocking_delay_DelayMs, _embedded_hal_blocking_delay_DelayUs,
};
use {defmt_rtt as _, panic_probe as _};

use {embassy_executor as executor, embassy_stm32 as stm32};

use ads1256::standard::input::{Differential, SingleEnded};
use ads1256::{
    AdControl, Ads1256, AutoCal, BitOrder, BlockingDelay, Buffer, ClockOut, Config, DState,
    DataRate, DigitalIo, DigitalIoDirection, DigitalIoState, DioDirection, Gain, Multiplexer,
    MuxInput, OutputPin, Sdcs, SpiBus, Status, Wait,
};
use embassy_time::{Delay, Duration, Timer};
use executor::Spawner;
use stm32::dma::NoDma;
use stm32::exti::ExtiInput;
use stm32::gpio::{Input, Level, Output, Pull, Speed};
use stm32::spi::Spi;
use stm32::time::Hertz;
use stm32::{pac, spi};

use uom::si::electric_potential::{millivolt, volt};
use uom::si::f32;

use defmt::{debug, error, info, trace, unwrap};
use physical::transducer::{lm35, thermocouple_k};
use uom::si::thermodynamic_temperature::degree_celsius;

const AUTOCAL_CONF: Config = Config {
    status: Status::setting(Buffer::Enabled, AutoCal::Enabled, BitOrder::MostSigFirst),
    multiplexer: Multiplexer::setting(MuxInput::AIn0, MuxInput::AIn1),
    ad_control: AdControl::setting(Gain::X64, Sdcs::Off, ClockOut::Off),
    data_rate: DataRate::Sps2_5,
    digital_io: DigitalIo::setting(DigitalIoState::default(), DigitalIoDirection::default()),
};

struct Ads1256Delay;

impl ads1256::BlockingDelay for Ads1256Delay {
    #[inline]
    fn t6_delay(&mut self) {
        Delay.delay_us(1u32);
    }

    fn t11_1_delay(&mut self) {}

    fn t11_2_delay(&mut self) {}
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    unsafe {
        pac::FLASH.acr().modify(|v| {
            v.set_prften(true);
            v.set_icen(true);
            v.set_dcen(true);
        });
    }

    let p = embassy_stm32::init(Default::default());

    let mut spi_conf = spi::Config::default();
    spi_conf.mode = spi::MODE_1;
    spi_conf.bit_order = spi::BitOrder::MsbFirst;

    let ads1256_data_ready = ExtiInput::new(Input::new(p.PF6, Pull::Up), p.EXTI6);
    let select_ads1256 = Output::new(p.PF7, Level::High, Speed::VeryHigh);

    let mut spi = Spi::new(
        p.SPI3,
        p.PC10,
        p.PC12,
        p.PC11,
        NoDma,
        NoDma,
        Hertz(ads1256::defaults::SPI_CLK_HZ),
        spi_conf,
    );

    let mut ads_1256 = Ads1256::new(Ads1256Delay, select_ads1256, ads1256_data_ready);
    ads_1256.write_config(&mut spi, AUTOCAL_CONF).unwrap();
    ads_1256.self_calibrate(&mut spi).await.unwrap();

    loop {
        let gain = Gain::X4;
        let reference = ads_1256
            .autocal_convert(
                &mut spi,
                SingleEnded::AIn2.into(),
                None,
                Some(AUTOCAL_CONF.ad_control.with_gain(gain)),
                None,
                false,
            )
            .await
            .unwrap()
            .to_voltage(gain);
        let reference = lm35::convert(reference).unwrap();
        let reference_celsius = reference.get::<degree_celsius>();
        info!("Reference junction temperature: {}°C", reference_celsius);

        let gain = Gain::X64;
        let voltage = ads_1256
            .autocal_convert(
                &mut spi,
                Differential::AIn0.into(),
                None,
                Some(AUTOCAL_CONF.ad_control.with_gain(gain)),
                None,
                false,
            )
            .await
            .unwrap()
            .to_voltage(AUTOCAL_CONF.ad_control.gain());
        let mv = voltage.get::<millivolt>();
        let temperature = thermocouple_k::convert_direct(voltage, reference).unwrap();
        let celsius = temperature.get::<degree_celsius>();
        info!("Thermocouple temperature: {}°C, millivolts: {}", celsius, mv);
    }
}
